from kain.classes import (
    Missing,
    Nothing,
)
from kain.descriptors import (
    cache,
    class_property,
    mixed_property,
    pin,
    proxy_to,
)
from kain.importer import (
    add_path,
    optional,
    required,
    sort,
)
from kain.internals import (
    Is,
    Who,
    to_ascii,
    to_bytes,
    unique,
)
from kain.monkey import (
    Monkey,
)
from kain.signals import (
    on_quit,
    quit_at,
)

__all__ = (
    'Is',
    'Missing',
    'Monkey',
    'Nothing',
    'Who',
    'add_path',
    'cache',
    'class_property',
    'mixed_property',
    'on_quit',
    'optional',
    'pin',
    'proxy_to',
    'quit_at',
    'required',
    'sort',
    'to_ascii',
    'to_bytes',
    'unique',
)



import sys
from collections import deque, namedtuple
from collections.abc import Callable, Collection, Iterable, Mapping, Sequence
from contextlib import suppress
from functools import cache, partial
from inspect import (
    getmodule,
    getsourcefile,
    isawaitable,
    isbuiltin,
    isclass,
    iscoroutine,
    isfunction,
    ismethod,
    ismodule,
    stack,
)
from itertools import filterfalse
from operator import itemgetter, methodcaller
from pathlib import Path
from platform import architecture
from re import sub
from sys import modules, stderr, stdin, stdout
from sysconfig import get_paths
from types import FunctionType, LambdaType, UnionType
from typing import Any, GenericAlias, get_args, get_origin

Collections = deque, dict, list, set, tuple, bytearray
Primitives = bool, float, int, str, complex, bytes
Builtins = Primitives + Collections

WinNT = 'windows' in architecture()[1].lower()

__all__ = (
    'Is',
    'Who',
    'get_attr',
    'get_owner',
    'iter_inheritance',
    'iter_stack',
    'to_ascii',
    'to_bytes',
    'unique',
)


def class_of(obj: Any) -> bool:
    return obj if isclass(obj) else type(obj)


def is_callable(obj: Any) -> bool:
    return isinstance(obj, Callable) or callable(obj)


def is_collection(obj: Any) -> bool:
    return (
        (
            isinstance(obj, Collection)
            and isinstance(obj, Sequence)
            and not isinstance(obj, bytes | str)
        )
        or is_mapping(obj)
        or all(map(partial(hasattr, obj), ('__getitem__', '__setitem__', '__delitem__')))
    )


def is_iterable(obj: Any) -> bool:
    return isinstance(obj, Iterable) or hasattr(obj, '__iter__')


def is_mapping(obj: Any) -> bool:
    return isinstance(obj, Mapping) or issubclass(class_of(obj), dict)


def is_primitive(obj: Any) -> bool:
    return obj is True or obj is False or obj is None or type(obj) in Builtins


def is_from_primivite(obj: Any) -> bool:
    return bool(obj is None or isinstance(obj, Primitives))


def is_from_builtin(obj: Any) -> bool:
    return bool(isinstance(obj, Collections) or is_from_primivite(obj))


def is_tty() -> bool:
    if not getattr(sys, 'frozen', False):  # nuitka compiler checks this
        return all(map(methodcaller('isatty'), (stderr, stdin, stdout)))
    return False


@cache
def _get_module_path_type(full: Any) -> tuple[bool | None, str]:
    dirs = get_paths()

    path = str(full)
    if WinNT:
        path = path.lower()

    for scheme, reason in (
        ('stdlib', True),
        ('purelib', False),
        ('platlib', False),
        ('platstdlib', True),
    ):
        subdir = dirs[scheme]
        if WinNT:
            subdir = subdir.lower()

        if path.startswith(subdir):
            return reason, str(full)[len(subdir) + 1 :]

    subdir = str(Path(__file__).parent.parent)

    if WinNT:
        subdir = subdir.lower()

    if path.startswith(subdir):
        return False, str(full)[len(subdir) + 1 :]

    return None, str(full)


def is_internal(x: Any) -> bool:
    if isbuiltin(x) or isbuiltin(class_of(x)):
        return True

    if module := get_module(x):

        if module.__name__ == 'builtins':
            return True

        is_stdlib = _get_module_path_type(module.__file__)[0]
        if is_stdlib is not None:
            return is_stdlib

    return False


def is_subclass(obj: Any, types) -> bool:  # noqa: PLR0911
    if types is None:
        return False

    if types in (Any, obj, object):
        return True

    if obj is None and types is class_of(None):
        return True

    cls = class_of(obj)

    if origin := get_origin(types):
        if cls is origin:
            return True

        args = get_args(types)

        if args and (Any in args or cls in args):
            # Any | None
            return True

        if origin is UnionType and (issubclass(cls, types) or (args and cls in args)):
            return True

        if class_of(types) is GenericAlias:
            # dict[str, str])
            return issubclass(cls, origin)

    return issubclass(cls, types)


def get_module(x: Any):
    if ismodule(x):
        return x

    if (module := getmodule(x)) or (module := getmodule(class_of(x))):
        return module


def get_module_name(x: Any) -> str | None:
    if module := get_module(x):
        with suppress(AttributeError):
            return module.__spec__.name


def object_name(obj: Any, full=True) -> str:
    def post(x):
        return sub(r'^([\?\.]+)', '', sub('^(__main__|__builtin__|builtins)', '', x))

    def get_module_from(x):
        return getattr(x, '__module__', get_module_name(x)) or '?'

    def get_object_name(x):
        if obj is Any:
            return 'typing.Any' if full else 'Any'

        name = getattr(x, '__qualname__', x.__name__)
        module = get_module_from(x)

        if not name.startswith(module):
            name = f'{module}.{name}'
        return name

    def main(obj):
        if ismodule(obj):
            return get_module_name(obj)

        for itis in iscoroutine, isfunction, ismethod:
            if itis(obj):
                name = get_object_name(obj)
                with suppress(AttributeError):
                    name = f'{object_name(obj.im_self or obj.im_class)}.{post(name)}'
                return name

        cls = class_of(obj)
        if cls is property:
            return get_object_name(obj.fget)

        return get_object_name(cls)

    name = post(main(obj))
    return name if full else name.rsplit('.', 1)[-1]


def pretty_module(obj: Any) -> str:
    return Who(obj).rsplit('.', 1)[0]


def source_file(obj: Any, template=None, **kw) -> str:
    kw.setdefault('exclude_self', False)
    kw.setdefault('exclude_stdlib', False)

    for child in iter_inheritance(class_of(obj), **kw):
        try:
            if path := getsourcefile(child):
                return (template % path) if template else str(path)
        except TypeError:  # noqa: PERF203
            ...


def just_value(obj: Any, /, **kw) -> str:
    kw.setdefault('addr', False)

    name = Who(obj, **kw)
    if isclass(obj):
        return f'({name})'
    return f'({name}){obj}'


def who_is(obj: Any, /, **kw) -> str:  # noqa: N802
    kw.setdefault('addr', True)
    return just_value(obj, **kw)


@cache
def is_imported_module(name: str) -> bool:

    with suppress(KeyError):
        return bool(modules[name])

    chunks = name.split('.')
    return (
        sum('.'.join(chunks[: no + 1]) in modules for no in range(len(chunks))) >= 2
    )  # noqa: PLR2004


def get_mro(obj, /, **kw):

    func = kw.pop('func', None)
    glue = kw.pop('glue', None)

    result = iter_inheritance(obj, **kw)

    if func:
        result = tuple(map(func, result))

    if glue:
        result = glue.join(result)

    return result if (func or glue) else tuple(result)


#


def simple_repr(x: Any) -> bool | str | None:
    if (x is None or x is True or x is False) or isinstance(x, str):
        return x

    if isinstance(x, int | float):
        return repr(x)

    return Who.Cast(x)


def format_args_and_keywords(*args, **kw) -> str:
    def format_args(x):
        return repr(tuple(map(simple_repr, x)))[1:-1].rstrip(',')

    def format_kwargs(x):
        return ', '.join(f'{k}={simple_repr(v)}' for k, v in x.items())

    if args and kw:
        return f'{format_args(args)}, {format_kwargs(kw)}'

    if args:
        return format_args(args)

    if kw:
        return format_kwargs(kw)

    return ''


# public interface, Is/Who


def Who(obj: Any, /, full=True, addr=False) -> str:  # noqa: N802
    key = '__name_full__' if full else '__name_short__'

    def get_name() -> str:
        try:
            store = obj.__dict__
            with suppress(KeyError):
                return store[key]
        except AttributeError:
            store = None

        name = object_name(obj, full=full)
        if store is not None:
            with suppress(AttributeError, TypeError):
                setattr(obj, key, name)
        return name

    name = get_name()
    if not addr:
        return name
    return f'{name}#{id(obj):x}'


Who.Args = format_args_and_keywords
Who.Cast = just_value
Who.File = source_file
Who.Inheritance = get_mro
Who.Is = who_is
Who.Module = pretty_module
Who.Name = partial(Who, full=False)

#

Is = namedtuple(
    'Is',
    (
        'Builtin '
        'Class '
        'Primivite '
        'tty '
        'awaitable '
        'builtin '
        'callable '
        'classOf '
        'collection '
        'coroutine '
        'function '
        'imported '
        'internal '
        'iterable '
        'mapping '
        'method '
        'module '
        'primitive '
        'subclass'
    ),
)(
    is_from_builtin,
    isclass,
    is_from_primivite,
    is_tty,
    isawaitable,
    isbuiltin,
    is_callable,
    class_of,
    is_collection,
    iscoroutine,
    isfunction,
    is_imported_module,
    is_internal,
    is_iterable,
    is_mapping,
    ismethod,
    ismodule,
    is_primitive,
    is_subclass,
)


# public functions


def iter_stack(*args, **kw):
    result = stack()[kw.pop('offset', 0) :]
    yield from (map(itemgetter(*args), result) if args else result)


def iter_inheritance(  # noqa: PLR0913
    obj: Any,
    include=None,
    exclude=None,
    exclude_self=True,
    exclude_stdlib=True,
    reverse=False,
):
    order = class_of(obj).__mro__[:-1]

    if not exclude_self:
        order = unique((obj, *order), key=id)
    else:
        order = unique(filter(lambda x: x is not obj, order), key=id)

    if reverse:
        order = reversed(list(order))

    if include:
        if isinstance(include, FunctionType | LambdaType):
            order = filter(include, order)
        else:
            if not is_iterable(include):
                include = (include,)
            order = filter(include.__contains__, order)

    if exclude:
        if isinstance(exclude, FunctionType | LambdaType):
            order = filterfalse(exclude, order)
        else:
            if not is_iterable(exclude):
                exclude = (exclude,)
            order = filterfalse(exclude.__contains__, order)

    if exclude_stdlib:
        order = filterfalse(is_internal, order)

    yield from order


def _get_attribute_from_inheritance(obj: Any, name: str, **kw) -> tuple[Any, Any]:

    index = kw.pop('index', 0)
    kw.setdefault('exclude_self', False)
    kw.setdefault('exclude_stdlib', False)

    counter = 0
    for child in iter_inheritance(obj, **kw):
        try:
            attr = child.__dict__[name]

        except KeyError:
            continue

        if not counter - index:
            return attr, child
        counter += 1

    raise KeyError(name)


def get_owner(obj: Any, name: str, **kw) -> Any:
    with suppress(KeyError):
        return _get_attribute_from_inheritance(obj, name, **kw)[1]


def get_attr(obj: Any, name: str, default: Any = None, **kw) -> Any:
    try:
        return _get_attribute_from_inheritance(obj, name, **kw)[0]
    except KeyError:
        return default


def to_ascii(x: str, /, charset: str | None = None) -> str:
    if not isinstance(x, bytes | str):
        raise TypeError(f'only bytes | str acceptable, not {Who.Cast(x)}')

    if isinstance(x, str):
        return x

    charset = charset or 'ascii'
    return to_bytes(x, charset=charset).decode(charset)


def to_bytes(x: str, /, charset: str | None = None) -> bytes:
    if not isinstance(x, bytes | str):
        raise TypeError(f'only bytes | str acceptable, not {Who.Cast(x)}')

    if not isinstance(x, str):
        return x

    return x.encode(charset or 'ascii')


def unique(iterable, /, key=None, include=None, exclude=None):
    skip = include is None

    if not key:
        exclude = set(exclude or ())
        include = frozenset(include or ())

    else:
        exclude = set(map(key, exclude or ()))
        include = frozenset(map(key, include or ()))

    excluded = exclude.__contains__
    included = include.__contains__
    is_dict = is_mapping(iterable)

    for element in iterable:

        k = key(element) if key else element
        if not excluded(k) and (skip or included(k)):

            yield (element, iterable[element]) if is_dict else element
            exclude.add(k)



from asyncio import ensure_future, iscoroutinefunction
from contextlib import suppress
from functools import cached_property, lru_cache, partial, wraps
from inspect import iscoroutine, isfunction, ismethod
from logging import getLogger
from operator import attrgetter
from time import time

from kain.classes import Missing
from kain.internals import (
    Is,
    Who,
    get_attr,
    get_owner,
)

__all__ = 'cache', 'class_property', 'mixed_property', 'pin', 'proxy_to'

Nothing = Missing()
logger = getLogger(__name__)


class PropertyError(Exception): ...


class ContextFaultError(PropertyError): ...


class ReadOnlyError(PropertyError): ...


class AttributeException(PropertyError): ...  # noqa: N818


def cache(limit=None):

    function = partial(lru_cache, maxsize=None, typed=False)

    if isinstance(limit, classmethod | staticmethod):
        msg = f"can't wrap {Who.Is(limit)}, you must use @cache after it"
        raise TypeError(msg)

    for func in isfunction, iscoroutine, ismethod:
        if func(limit):
            return function()(limit)

    if limit is not None and (not isinstance(limit, float | int) or limit <= 0):
        msg = f'limit must be None or positive integer, not {Who.Is(limit)}'
        raise TypeError(msg)

    return function(maxsize=limit) if limit else function()


def extract_wrapped(desc):
    # when it's default instance-method replacer
    if Is.subclass(desc, InsteadProperty):
        return desc.__get__

    # when it's full-featured (cached) property
    if Is.subclass(desc, AbstractProperty):
        return desc.call

    # when it's builtin @property
    if Is.subclass(desc, property):
        return desc.fget

    # when wrapped functions stored in .func
    if Is.subclass(desc, cached_property):
        return desc.func

    msg = (
        f"couldn't extract wrapped function from {Who(desc)}: "
        f"replace it with @property, @cached_property, @{Who(pin)}, "
        f"or other descriptor derived from {Who(AbstractProperty)}"
    )
    raise NotImplementedError(msg)


def parent_call(func):

    @wraps(func)
    def parent_caller(node, *args, **kw):
        try:
            desc = get_attr(
                Is.classOf(node),
                func.__name__,
                exclude_self=True,
                index=func.__name__ not in Is.classOf(node).__dict__,
            )

            return func(node, extract_wrapped(desc)(node, *args, **kw), *args, **kw)

        except RecursionError as e:
            msg = (
                f"{Who(node)}.{func.__name__} call real {Who(func)}, "
                f"couldn't reach parent descriptor; "
                f"maybe {Who(func)} it's mixin of {Who(node)}?"
            )
            raise RecursionError(msg) from e

    return parent_caller


def invokation_context_check(func):

    @wraps(func)
    def context(self, node, *args, **kw):
        if (klass := self.klass) is not None and (
            node is None or klass != Is.Class(node)
        ):
            msg = f'{Who(func)} exception, {self.header_with_context(node)}, {node=}'

            if node is None and not klass:
                msg = f'{msg}; looks like as non-instance invokation'
            raise ContextFaultError(msg)

        return func(self, node, *args, **kw)

    return context


class AbstractProperty:
    @classmethod
    def with_parent(cls, function):
        return cls(parent_call(function))

    def __init__(self, function):
        self.function = function

    @cached_property
    def name(self):
        return self.function.__name__

    @property
    def title(self):
        raise NotImplementedError

    @cached_property
    def header(self):
        try:
            return f'{self.title}({self.function!a})'
        except Exception:  # noqa: BLE001
            return f'{self.title}({Who(self.function)})'

    def header_with_context(self, node):
        raise NotImplementedError


class CustomCallbackMixin:
    @classmethod
    def by(cls, callback):
        if not Is.subclass(cls, Cached):
            cls = Cached  # noqa: PLW0642
        return partial(cls, is_actual=callback)

    expired_by = by

    @classmethod
    def ttl(cls, expire: float):
        if not isinstance(expire, float | int):
            msg = f'expire must be float or int, not {Who.Cast(expire)}'
            raise TypeError(msg)

        if expire <= 0:
            msg = f'expire must be positive number, not {expire!r}'
            raise ValueError(msg)

        def is_actual(self, node, value=Nothing):  # noqa: ARG001
            return (value + expire > time()) if value else time()

        return cls.by(is_actual)


class InsteadProperty(AbstractProperty, CustomCallbackMixin):
    def __init__(self, function):
        if iscoroutinefunction(function):
            msg = (
                f'{Who(function)} is coroutine function, '
                'you must use @pin.native instead of just @pin'
            )
            raise TypeError(msg)
        super().__init__(function)

    @cached_property
    def title(self):
        return f'instance just-replace-descriptor {Who(self, addr=True)}'

    def header_with_context(self, node):
        return (
            f'{self.header} called with '
            f'{("instance", "class")[Is.Class(node)]} '
            f'({Who(node, addr=True)})'
        )

    def __get__(self, node, klass=Nothing):
        if node is None:
            raise ContextFaultError(self.header_with_context(klass))

        with suppress(KeyError):
            return node.__dict__[self.name]

        value = self.function(node)
        node.__dict__[self.name] = value
        return value

    def __delete__(self, node):
        msg = f'{self.header_with_context(node)}: deleter called'
        raise ReadOnlyError(msg)


class BaseProperty(AbstractProperty):

    klass = False
    readonly = False

    @InsteadProperty
    def is_data(self):
        return bool(hasattr(self, '__set__') or hasattr(self, '__delete__'))

    @InsteadProperty
    def title(self):
        mode = 'mixed' if self.klass is None else ('instance', 'class')[self.klass]

        prefix = ('', 'data ')[self.is_data]
        return f'{mode} {prefix}descriptor {Who(self, addr=True)}'.strip()

    def header_with_context(self, node):
        if node is None:
            mode = 'mixed' if self.klass is None else 'undefined'
        else:
            mode = ('instance', 'class')[Is.Class(node)]
        return (
            f'{self.header} with {mode} type called with'
            f'{("instance", "class")[Is.Class(node)]} '
            f'({Who(node, addr=True)})'
        )

    @invokation_context_check
    def get_node(self, node):
        return node

    @invokation_context_check
    def call(self, node):
        try:
            value = self.function(node)
            if not iscoroutinefunction(self.function):
                return value
            return ensure_future(value)

        except AttributeError as e:
            error = AttributeException(str(e).rsplit(':', 1)[-1])

            error.exception = e
            raise error from e

    def __str__(self):
        return f'<{self.header}>'

    def __repr__(self):
        return f'<{self.title}>'

    def __get__(self, instance, klass):
        if instance is None and self.klass is False:
            raise ContextFaultError(self.header_with_context(klass))

        return self.call((instance, klass)[self.klass])


class InheritedClass(BaseProperty):
    """By default class property will be used parent class.
    This class change behavior to last inherited child.
    """

    @invokation_context_check
    def get_node(self, node):
        return node

    @classmethod
    def make_from(cls, parent):
        """Make child-aware class from plain parent-based."""

        name = Who(parent, full=False)
        suffix = f'{"_" if name == name.lower() else ""}inherited'.capitalize()

        result = type(f'{name}{suffix}', (cls, parent), {})
        result.here = parent
        return result


class Cached(BaseProperty, CustomCallbackMixin):

    def __init__(self, function, is_actual=Nothing):
        super().__init__(function)

        if method := getattr(Is.classOf(self), 'is_actual', None):
            if is_actual:
                msg = (
                    f"{Who.Is(self)}.is_actual method ({Who.Cast(method)}) "
                    f"can't override by is_actual kw: {Who.Cast(is_actual)}"
                )
                raise TypeError(msg)
            is_actual = method
        self.is_actual = is_actual

    @invokation_context_check
    def get_cache(self, node):
        name = f'__{("instance", "class")[Is.Class(node)]}_memoized__'

        if hasattr(node, '__dict__'):
            with suppress(KeyError):
                return node.__dict__[name]

        cache = {}
        setattr(node, name, cache)
        return cache

    @invokation_context_check
    def call(self, obj):
        node = self.get_node(obj)
        with suppress(KeyError):

            stored = self.get_cache(node)[self.name]
            if not self.is_actual:
                return stored

            value, stamp = stored
            if self.is_actual(self, node, stamp) is True:
                return value

        return self.__set__(node, super().call(obj))

    @invokation_context_check
    def __set__(self, node, value):
        cache = self.get_cache(node)

        if not self.is_actual:
            cache[self.name] = value
        else:
            cache[self.name] = value, self.is_actual(self, node)
        return value

    @invokation_context_check
    def __delete__(self, node):
        cache = self.get_cache(node)
        with suppress(KeyError):
            del cache[self.name]


class ClassProperty(BaseProperty):
    klass = True

    @invokation_context_check
    def get_node(self, node):
        return get_owner(node, self.name) if Is.Class(node) else node


class MixedProperty(ClassProperty):
    klass = None

    def __get__(self, instance, klass):
        return self.call(instance or klass)


class ClassCachedProperty(ClassProperty, Cached):
    """Class-level cached property that passes the original class as the first
    positional argument and replaces the original data-descriptor."""


class MixedCachedProperty(MixedProperty, Cached):
    """Mixed-level cached property that replaces the original data-descriptor"""


class PreCachedProperty(MixedProperty, Cached):

    @invokation_context_check
    def __set__(self, node, value):
        if not Is.Class(node):
            return value
        return super().__set__(node, value)


class PostCachedProperty(MixedProperty, Cached):

    @invokation_context_check
    def __set__(self, node, value):
        if Is.Class(node):
            return value
        return super().__set__(node, value)


#


class pin(InsteadProperty):  # noqa: N801

    native = Cached
    cls = InheritedClass.make_from(ClassCachedProperty)
    any = InheritedClass.make_from(MixedCachedProperty)
    pre = InheritedClass.make_from(PreCachedProperty)
    post = InheritedClass.make_from(PostCachedProperty)


class class_property(ClassProperty): ...  # noqa: N801


class mixed_property(MixedProperty): ...  # noqa: N801


#


def proxy_to(  # noqa: PLR0915
    *mapping,
    getter=attrgetter,
    default=Nothing,
    pre=None,
    safe=True,
):
    if isinstance(mapping[-1], str):
        bind = pin

    elif mapping[-1] is None:
        bind, mapping = None, mapping[:-1]

    else:
        bind, mapping = mapping[-1], mapping[:-1]

    def class_wraper(cls):  # noqa: PLR0915
        if not Is.Class(cls):
            msg = f"{Who.Is(cls)} isn't a class"
            raise TypeError(msg)

        try:
            fields = cls.__proxy_fields__
        except AttributeError:
            fields = []
            cls.__proxy_fields__ = fields

        pivot, mapping_list = mapping[0], mapping[1:]

        if not mapping_list or (
            len(mapping_list) == 1 and not isinstance(mapping_list[0], str)
        ):
            raise ValueError(f'empty {mapping_list=} for {pivot=}')

        for method in mapping_list:

            if safe and not method.startswith('_') and get_attr(cls, method):
                msg = (
                    f'{Who(cls)} already exists {method!a}: {get_attr(cls, method)}'
                )
                raise TypeError(msg)

            def wrapper(name, node):
                if not isinstance(pivot, str):
                    try:
                        return getattr(pivot, name)
                    except AttributeError as e:
                        msg = (
                            f"{Who(node)}.{name} {Who.Name(getter)[:4]}-proxied -> "
                            f"{Who(pivot)}.{name}, but last isn't exists"
                        )
                        raise AttributeError(msg) from e

                try:
                    entity = getattr(node, pivot)
                except AttributeError as e:
                    msg = (
                        f"{Who(node)}.{name} {Who.Name(getter)[:4]}-proxied -> "
                        f"{Who(node)}.{pivot}.{name}, but "
                        f"{Who(node)}.{pivot} isn't exists"
                    )
                    raise AttributeError(msg) from e

                if entity is None:
                    msg = (
                        f'{Who(node)}.{name} {Who.Name(getter)[:4]}-proxied -> '
                        f'{Who(node)}.{pivot}.{name}, but current '
                        f'{Who(node)}.{pivot} is None'
                    )

                    if default is Nothing:
                        raise AttributeError(msg)

                    msg = f'{msg}; return {Who.Is(default)}'
                    logger.warning(msg)
                    result = default

                else:
                    try:
                        result = getter(name)(entity)

                    except (AttributeError, KeyError) as e:
                        msg = (
                            f"{Who(node)}.{name} {Who.Name(getter)[:4]}-proxied -> "
                            f"{Who(node)}.{pivot}.{name}, but isn't exists "
                            f"('{name}' not in {Who(node)}.{pivot}): "
                            f"{Who.Is(entity)}"
                        )

                        if default is Nothing:
                            raise Is.classOf(e)(msg) from e

                        msg = f'{msg}; return {Who.Is(default)}'
                        logger.warning(msg)
                        result = default

                return partial(pre, result) if pre else result

            wrapper.__name__ = method
            wrapper.__qualname__ = f'{pivot}.{method}'

            if bind is None:
                node = cls.__dict__[pivot]
                try:
                    value = node.__dict__[method]
                except KeyError:
                    value = getattr(node, method)
            else:
                wrap = partial(wrapper, method)
                wrap.__name__ = method
                wrap.__qualname__ = f'{pivot}.{method}'
                value = bind(wrap)

            fields.append(method)
            setattr(cls, method, value)
            cls.__proxy_fields__.sort()

        return cls

    return class_wraper
